
# This file was generated by 'versioneer.py' (0.14) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.35.0+10.g143f70e'
version_full = '143f70e9005ed5ebc8d153d621d221d9aad19448'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

